from ipaddress import ip_interface
from pathlib import Path
from typing import Literal, Optional
import json
import logging

import click
from pydantic import (
    Field,
    IPvAnyInterface,
    IPvAnyAddress,
    constr,
    SecretStr,
    PrivateAttr,
)
from ruamel.yaml import YAML

from .paths import get_config_path, get_secret_path
from .utils import (
    StrictModel,
    check_file_mode,
    check_key,
    check_key_pair,
    ensure_file,
)
from .wg import gen_key_pair, genpsk
from .wg_quick import WgQuickConfig, WgQuickInterfaceConfig, WgQuickPeerConfig

logger = logging.getLogger(__name__)


class WgWizardPeerConfig(StrictModel):
    listen_port: Optional[int]
    fw_mark: Optional[Literal["off"] | int]
    addresses: list[IPvAnyInterface] = Field(min_items=1)
    dns_addresses: list[IPvAnyAddress] = Field(default_factory=list)
    mtu: Optional[int]
    table: Optional[str]
    pre_up: list[str] = Field(default_factory=list)
    post_up: list[str] = Field(default_factory=list)
    pre_down: list[str] = Field(default_factory=list)
    post_down: list[str] = Field(default_factory=list)
    server_allowed_ips: list[IPvAnyInterface] = Field(min_items=1)
    server_endpoint: Optional[constr(regex=r".+:\d+")]  # noqa: F722
    server_persistent_keepalive: Optional[Literal["off"] | int]
    client_allowed_ips: list[IPvAnyInterface] = Field(min_items=1)
    client_endpoint: Optional[constr(regex=r".+:\d+")]  # noqa: F722
    client_persistent_keepalive: Optional[Literal["off"] | int]


class WgWizardConfig(StrictModel):
    name: constr(regex=r"[a-zA-Z0-9_=+.-]{1,15}")  # noqa: F722
    listen_port: int
    fw_mark: Optional[Literal["off"] | int]
    addresses: list[IPvAnyInterface] = Field(min_items=1)
    # TODO: test DNS DHCP
    dns_addresses: list[IPvAnyAddress] = Field(default_factory=list)
    mtu: Optional[int]
    table: Optional[str]
    pre_up: list[str] = Field(default_factory=list)
    post_up: list[str] = Field(default_factory=list)
    pre_down: list[str] = Field(default_factory=list)
    post_down: list[str] = Field(default_factory=list)
    default_endpoint: constr(regex=r".+:\d+")  # noqa: F722
    peers: dict[
        constr(regex=r"[a-zA-Z0-9_=+.-]+"), WgWizardPeerConfig  # noqa: F722
    ] = Field(default_factory=dict)
    _yaml: dict = PrivateAttr(default=None)

    @classmethod
    def from_file(cls, path: Path):
        yaml = YAML()
        raw_config = yaml.load(path)
        config = cls(**raw_config)
        config._yaml = raw_config
        return config

    def dump(self, path: Path, overwrite=False):
        path = path.resolve()
        logger.info("Writing config to %s", path)
        ensure_file(path, mode=0o600, overwrite=overwrite)
        yaml = YAML()
        yaml.indent(mapping=2, sequence=4, offset=2)
        if self._yaml is None:
            yaml.dump(json.loads(self.json(exclude_unset=True)), path)
        else:
            yaml.dump(self._yaml, path)

    def find_next_available_interface(self) -> Optional[IPvAnyInterface]:
        existing_ips = set(address.ip for address in self.addresses)
        existing_ips.update(
            address.ip
            for _peer_config in self.peers.values()
            for address in _peer_config.addresses
        )
        for _address in self.addresses:
            for ip in _address.network.hosts():
                if ip not in existing_ips:
                    return ip_interface(ip)
        return None

    def add_peer(self, name: str, peer_config: WgWizardPeerConfig):
        self.peers[name] = peer_config
        if self._yaml is not None:
            if "peers" not in self._yaml:
                self._yaml["peers"] = {}
            self._yaml["peers"][name] = json.loads(peer_config.json(exclude_unset=True))


class WgWizardPeerSecret(StrictModel):
    private_key: SecretStr
    public_key: constr(min_length=1)
    preshared_key: Optional[SecretStr]

    @classmethod
    def generate(cls) -> "WgWizardPeerSecret":
        private_key, public_key = gen_key_pair()
        preshared_key = genpsk()
        return cls(
            private_key=private_key,
            public_key=public_key,
            preshared_key=preshared_key,
        )

    def check(self, name: str):
        check_key_pair(
            self.private_key.get_secret_value(), self.public_key, f"Peer {name}"
        )
        if self.preshared_key is not None:
            check_key(
                self.preshared_key.get_secret_value(), f"Peer {name} preshared_key"
            )


class WgWizardSecret(StrictModel):
    private_key: SecretStr
    public_key: constr(min_length=1)
    peers: dict[str, WgWizardPeerSecret] = Field(default_factory=dict)

    @classmethod
    def generate(cls) -> "WgWizardSecret":
        private_key, public_key = gen_key_pair()
        return cls(private_key=private_key, public_key=public_key)

    def regenerate_server_secret(self):
        self.private_key, self.public_key = gen_key_pair()

    @classmethod
    def from_file(cls, path: Path):
        check_file_mode(path)
        return cls.parse_file(path)

    def dump(self, path: Path, overwrite=False):
        path = path.resolve()
        logger.info("Writing secret to %s", path)
        ensure_file(path, mode=0o600, overwrite=overwrite)
        path.write_text(self.json(indent=2))

    def generate_peer_secret(self, name: str) -> WgWizardPeerSecret:
        peer_secret = WgWizardPeerSecret.generate()
        self.peers[name] = peer_secret
        return peer_secret

    def check(self):
        check_key_pair(self.private_key.get_secret_value(), self.public_key, "Server")
        for peer_name, peer_secret in self.peers.items():
            peer_secret.check(peer_name)


class WgWizard(StrictModel):
    config: WgWizardConfig
    secret: WgWizardSecret

    @classmethod
    def from_dir(cls, config_dir, interface):
        config = WgWizardConfig.from_file(get_config_path(config_dir, interface))
        secret = WgWizardSecret.from_file(get_secret_path(config_dir, interface))
        return cls(config=config, secret=secret)

    def dump(self, config_dir, interface, overwrite=True, config=True, secret=True):
        if config:
            self.config.dump(get_config_path(config_dir, interface), overwrite)
        if secret:
            self.secret.dump(get_secret_path(config_dir, interface), overwrite)

    def check_secret(self):
        config_peers = set(self.config.peers.keys())
        secret_peers = set(self.secret.peers.keys())
        missing_secrets = config_peers - secret_peers
        redundant_secrets = secret_peers - config_peers
        if missing_secrets:
            raise ValueError(
                f"Peers missing secrets: {missing_secrets}. "
                "You can remove the redundant peers in the config or "
                "generate the secrets using `wg-wizard generate-keys --missing`."
            )
        if redundant_secrets:
            logger.warning("Redundant peers in the secret: %s", redundant_secrets)

        self.secret.check()

    def generate_keys(
        self,
        regenerate_all: bool = False,
        server: bool = False,
        missing: bool = False,
        peers: Optional[list[str]] = None,
        overwrite: bool = True,
    ) -> list[str]:
        if regenerate_all:
            if not overwrite:
                click.confirm(
                    (
                        "All the keys will be regenerated and cannot be reverted. "
                        "Are you sure you want to do this?"
                    ),
                    abort=True,
                )
            self.secret.regenerate_server_secret()
            self.secret.peers = {}
            for peer_name in self.config.peers:
                self.secret.generate_peer_secret(peer_name)
            return []
        if server:
            if not overwrite:
                click.confirm(
                    (
                        "The server secret will be regenerated and cannot be reverted. "
                        "Are you sure you want to do this?"
                    ),
                    abort=True,
                )
            self.secret.regenerate_server_secret()
        for peer_name in peers:
            if not overwrite and peer_name in self.secret.peers:
                click.confirm(
                    (
                        f"The secret for peer '{peer_name}' will be regenerated "
                        "and cannot be reverted. Are you sure you want to do this?"
                    ),
                    abort=True,
                )
            self.secret.generate_peer_secret(peer_name)
        missing_peers = []
        if missing:
            for peer_name in self.config.peers:
                if peer_name not in self.secret.peers:
                    self.secret.generate_peer_secret(peer_name)
                    missing_peers.append(peer_name)
        return missing_peers


def to_wg_quick_server_config(wg_wizard: WgWizard) -> WgQuickConfig:
    wg_wizard.check_secret()
    config = wg_wizard.config
    secret = wg_wizard.secret
    interface_config = WgQuickInterfaceConfig(
        private_key=secret.private_key,
        listen_port=config.listen_port,
        fw_mark=config.fw_mark,
        address=config.addresses,
        dns=config.dns_addresses,
        mtu=config.mtu,
        table=config.table,
        pre_up=config.pre_up,
        post_up=config.post_up,
        pre_down=config.pre_down,
        post_down=config.post_down,
    )
    peer_configs = [
        WgQuickPeerConfig(
            comment=peer_name,
            public_key=secret.peers[peer_name].public_key,
            preshared_key=secret.peers[peer_name].preshared_key,
            allowed_ips=peer_config.server_allowed_ips,
            endpoint=peer_config.server_endpoint,
            persistent_keepalive=peer_config.server_persistent_keepalive,
        )
        for peer_name, peer_config in config.peers.items()
    ]
    return WgQuickConfig(interface=interface_config, peer=peer_configs)


def to_wg_quick_client_config(wg_wizard: WgWizard, peer_name: str) -> WgQuickConfig:
    wg_wizard.check_secret()
    peer_config = wg_wizard.config.peers[peer_name]
    secret = wg_wizard.secret
    interface_config = WgQuickInterfaceConfig(
        private_key=secret.peers[peer_name].private_key,
        listen_port=peer_config.listen_port,
        fw_mark=peer_config.fw_mark,
        address=peer_config.addresses,
        dns=peer_config.dns_addresses,
        mtu=peer_config.mtu,
        table=peer_config.table,
        pre_up=peer_config.pre_up,
        post_up=peer_config.post_up,
        pre_down=peer_config.pre_down,
        post_down=peer_config.post_down,
    )
    peer_configs = [
        WgQuickPeerConfig(
            public_key=secret.public_key,
            preshared_key=secret.peers[peer_name].preshared_key,
            allowed_ips=peer_config.client_allowed_ips,
            endpoint=peer_config.client_endpoint or wg_wizard.config.default_endpoint,
            persistent_keepalive=peer_config.client_persistent_keepalive,
        )
    ]
    return WgQuickConfig(interface=interface_config, peer=peer_configs)


def export_wg_quick_config(
    wg_wizard: WgWizard,
    text: bool,
    qrcode: bool,
    invert_qrcode: bool,
    peer_name: Optional[str],
):
    if peer_name is None:
        server_config = to_wg_quick_server_config(wg_wizard)
    else:
        server_config = to_wg_quick_client_config(wg_wizard, peer_name)
    ini_str = server_config.format_ini()
    if text:
        print(ini_str)
    if qrcode:
        from qrcode import QRCode

        qr = QRCode()
        qr.add_data(ini_str)
        qr.print_ascii(invert=invert_qrcode)


def export_wg_quick_config_from_files(
    config_dir: str,
    interface: str,
    text: bool,
    qrcode: bool,
    invert_qrcode: bool,
    peer_name: Optional[str] = None,
):
    wg_wizard = WgWizard.from_dir(config_dir, interface)
    export_wg_quick_config(wg_wizard, text, qrcode, invert_qrcode, peer_name)



from base64 import standard_b64decode
import binascii
from email.policy import strict
from ipaddress import ip_interface
from pathlib import Path
import shlex
import stat

import click
from pydantic import BaseModel, SecretStr, Extra

from .wg import pubkey


def to_camel(string: str) -> str:
    return "".join(word.capitalize() for word in string.split("_"))


class StrictModel(BaseModel):
    class Config:
        extra = Extra.forbid
        validate_assignment = True
        validate_all = True
        json_encoders = {
            SecretStr: lambda v: v.get_secret_value(),
        }


class StrictCamelModel(StrictModel):
    class Config:
        alias_generator = to_camel
        allow_population_by_field_name = True


def ensure_file(path: Path, mode: int, overwrite=False):
    if not path.exists():
        path.touch(mode=mode)
    else:
        if not path.is_file():
            raise ValueError(f"{path} is not a file.")
        if not overwrite:
            click.confirm(
                f"'{path}' already exists. Do you want to overwrite it?", abort=True
            )
        path.chmod(mode=mode)


def check_file_mode(path: Path):
    st_mode = path.stat().st_mode
    if st_mode & (stat.S_IRWXG | stat.S_IRWXO):
        raise PermissionError(
            f"Permissions 0o{stat.S_IMODE(st_mode):o} for '{path}' are too open. "
            "Setting the mode to 0o600 is recommended."
        )


def format_ini_lines(obj: BaseModel, exclude=None):
    for field_name, field_config in obj.__fields__.items():
        if exclude is not None and field_name in exclude:
            continue
        field_val = getattr(obj, field_name)
        if field_val is None:
            continue
        if isinstance(field_val, list):
            for val in field_val:
                yield f"{field_config.alias} = {val}"
        elif isinstance(field_val, SecretStr):
            yield f"{field_config.alias} = {field_val.get_secret_value()}"
        elif isinstance(field_val, bool):
            yield f"{field_config.alias} = {'true' if field_val else 'false'}"
        else:
            yield f"{field_config.alias} = {field_val}"


def check_key(key: str, error_prefix: strict):
    try:
        decoded = standard_b64decode(key)
    except binascii.Error as exc:
        raise binascii.Error(f"{error_prefix} base64 failed decoding: {exc}")
    if len(decoded) != 32:
        raise ValueError(f"{error_prefix} size is not 32.")


def check_key_pair(private_key: str, public_key: str, error_prefix: str):
    check_key(private_key, f"{error_prefix} private_key")
    check_key(public_key, f"{error_prefix} public_key")
    if pubkey(private_key) != public_key:
        raise ValueError(f"{error_prefix}'s private_key and public_key is not a pair.")


def get_iptables_commands(
    internet_interface_name: str,
    allow_intranet: bool,
    allow_all_server_ip: bool,
    wg_server_interfaces: list[str],
) -> tuple[list[str], list[str]]:
    wg_server_interfaces = [
        ip_interface(interface) for interface in wg_server_interfaces
    ]

    post_up = []
    if internet_interface_name:
        ii_name = shlex.quote(internet_interface_name)
        post_up.append(f"iptables -A FORWARD -i %i -o {ii_name} -j ACCEPT")
        post_up.append(f"iptables -A FORWARD -i {ii_name} -o %i -j ACCEPT")
        for network in set(interface.network for interface in wg_server_interfaces):
            network = shlex.quote(str(network))
            post_up.append(
                f"iptables -t nat -A POSTROUTING -s {network} -o {ii_name} -j MASQUERADE"
            )
    if allow_intranet:
        post_up.append("iptables -A FORWARD -i %i -o %i -j ACCEPT")
    if not allow_all_server_ip:
        for interface in wg_server_interfaces:
            ip = shlex.quote(str(interface.ip))
            post_up.append(f"iptables -A INPUT -d {ip} -i %i -j ACCEPT")
        post_up.append("iptables -A INPUT -i %i -j DROP")

    pre_down = [cmd.replace(" -A ", " -D ", 1) for cmd in post_up]
    return post_up, pre_down



import subprocess


def genkey():
    return subprocess.run(
        ["wg", "genkey"], stdout=subprocess.PIPE, check=True, text=True
    ).stdout.strip()


def pubkey(private_key):
    return subprocess.run(
        ["wg", "pubkey"],
        input=private_key,
        stdout=subprocess.PIPE,
        check=True,
        text=True,
    ).stdout.strip()


def genpsk():
    return subprocess.run(
        ["wg", "genpsk"], stdout=subprocess.PIPE, check=True, text=True
    ).stdout.strip()


def gen_key_pair():
    private_key = genkey()
    public_key = pubkey(private_key)
    return private_key, public_key



from itertools import chain
from typing import Literal, Optional

from pydantic import Field, IPvAnyInterface, IPvAnyAddress, SecretStr

from .utils import StrictCamelModel, format_ini_lines


class WgQuickInterfaceConfig(StrictCamelModel):
    private_key: SecretStr
    listen_port: Optional[int]
    fw_mark: Optional[Literal["off"] | int]
    address: list[IPvAnyInterface] = Field(min_items=1)
    dns: list[IPvAnyAddress] = Field(alias="DNS")
    mtu: Optional[int] = Field(alias="MTU")
    table: Optional[str]
    pre_up: list[str]
    post_up: list[str]
    pre_down: list[str]
    post_down: list[str]
    save_config: Optional[bool]

    def format_ini_lines(self) -> list[str]:
        yield "[Interface]"
        yield from format_ini_lines(self)


class WgQuickPeerConfig(StrictCamelModel):
    comment: Optional[str]
    public_key: str
    preshared_key: Optional[SecretStr]
    allowed_ips: list[IPvAnyInterface] = Field(alias="AllowedIPs", min_items=1)
    endpoint: Optional[str]
    persistent_keepalive: Optional[Literal["off"] | int]

    def format_ini_lines(self) -> list[str]:
        yield "[Peer]"
        if self.comment is not None:
            yield f"# {self.comment}"
        yield from format_ini_lines(self, exclude={"comment"})


class WgQuickConfig(StrictCamelModel):
    interface: WgQuickInterfaceConfig
    peer: list[WgQuickPeerConfig]

    def format_ini(self) -> str:
        def check(s):
            assert "\n" not in s
            return s

        ini_str = "\n".join(
            check(s)
            for s in chain(
                self.interface.format_ini_lines(),
                *(chain(("",), peer.format_ini_lines()) for peer in self.peer),
            )
        )
        return ini_str



from pathlib import Path


def get_config_path(config_dir, interface) -> Path:
    return Path(config_dir, f"{interface}.yml")


def get_secret_path(config_dir, interface) -> Path:
    return Path(config_dir, f"{interface}_secret.json")



from functools import partial
import logging

import click

logger = logging.getLogger(__name__)
option = partial(click.option, show_default=True)
interface_option = option(
    "--interface",
    "-i",
    prompt="Interface name for WireGuard",
    help="Interface name for WireGuard.",
    default="wg0",
)
config_dir_option = option(
    "--config-dir",
    "-c",
    help="The directory for the wg-wizard configs and secrets.",
    type=click.Path(),
    default=".",
)
invert_qrcode_option = option(
    "--invert-qrcode/--no-invert-qrcode",
    is_flag=True,
    default=True,
    help="Whether to invert the color of the QR Code.",
)


@click.group(context_settings={"max_content_width": 120})
@click.version_option()
def main():
    logging.basicConfig(
        level=logging.INFO, format="[%(asctime)s][%(levelname)s] %(name)s: %(message)s"
    )


@main.command()
@interface_option
@config_dir_option
@option(
    "--listen-port",
    "-p",
    prompt="Interface.ListenPort of the relay server",
    help="Interface.ListenPort of the relay server.",
    default=51820,
    type=int,
)
@option(
    "--addresses",
    "-a",
    prompt="Interface.Address of the relay server",
    help="Interface.Address of the relay server.",
    default="192.168.10.1/24",
)
@option(
    "--default-endpoint",
    "-e",
    prompt="The default endpoint in clients' Peer.Endpoint configs (e.g., example.com:51280)",
    help="""
        The default endpoint in clients' Peer.Endpoint configs (e.g., example.com:51280).
        If the port is not provided, it will be added automatically according to the listen_port.
        The endpoint can also be overridden in peer configs.
    """,
)
@option(
    "--internet-interface-name",
    default="",
    prompt=(
        "If you want to allow the clients to access the internet via the relay server, "
        "you must provide the interface name you want to forward the internet traffic to. "
        "It's usually eth0 or wlan0. You can check it by executing `ip addr`. "
        "If you provide an interface name {interface}, the following rules will be added:\n"
        "- iptables -A FORWARD -i %i -o {interface} -j ACCEPT\n"
        "- iptables -A FORWARD -i {interface} -o %i -j ACCEPT\n"
        "- iptables -t nat -A POSTROUTING -s {network} -o {interface} -j MASQUERADE\n"
        "Interface name for connecting to the internet"
    ),
    help="""
        Interface name for connecting to the internet.
        If you want to allow the clients to access the internet via the relay server,
        you must provide the interface name you want to forward the internet traffic to.
        It's usually eth0 or wlan0. You can check it by executing `ip addr`.
        If you provide an interface name {interface}, the following rules will be added:
        [`iptables -A FORWARD -i %i -o {interface} -j ACCEPT`,
        `iptables -A FORWARD -i {interface} -o %i -j ACCEPT`,
        `iptables -t nat -A POSTROUTING -s {network} -o {interface} -j MASQUERADE`].
    """,
)
@option(
    "--allow-intranet/--no-allow-intranet",
    is_flag=True,
    default=True,
    prompt=(
        "Do you want to allow the clients to connect with each other? "
        "If yes, a rule will be added: `iptables -A FORWARD -i %i -o %i -j ACCEPT`"
    ),
    help="""
        Whether to allow the clients to connect with each other.
        If yes, a rule will be added: `iptables -A FORWARD -i %i -o %i -j ACCEPT`.
    """,
)
@option(
    "--allow-all-server-ip/--no-allow-all-server-ip",
    is_flag=True,
    default=False,
    prompt=(
        "Do you want to allow the clients to connect to any IPs on the relay server? "
        "If no, only the IP of the WireGuard interface can be connected, "
        "that is, the following rules will be added:\n"
        "- iptables -A INPUT -d {wg_server_interface_ip} -i %i -j ACCEPT\n"
        "- iptables -A INPUT -i %i -j DROP\n"
    ),
    help="""
        Wether to allow the clients to connect to any IPs on the relay server.
        If no, only the IP of the WireGuard interface can be connected,
        that is, the following rules will be added:
        [`iptables -A INPUT -d {wg_server_interface_ip} -i %i -j ACCEPT`,
        `iptables -A INPUT -i %i -j DROP`].
    """,
)
@option(
    "--overwrite/--no-overwrite",
    help="Whether to overwrite the config and secret files without confirmation.",
    is_flag=True,
)
def init(
    interface,
    config_dir,
    listen_port,
    addresses,
    default_endpoint,
    internet_interface_name,
    allow_intranet,
    allow_all_server_ip,
    overwrite,
):
    """Initialize a new wg-wizard config."""
    from .core import WgWizard, WgWizardConfig, WgWizardSecret
    from .utils import get_iptables_commands

    if ":" not in default_endpoint:
        default_endpoint += f":{listen_port}"
    addresses = [ip.strip() for ip in addresses.split(",")]

    post_up, pre_down = get_iptables_commands(
        internet_interface_name,
        allow_intranet,
        allow_all_server_ip,
        wg_server_interfaces=addresses,
    )
    config = WgWizardConfig(
        name=interface,
        listen_port=listen_port,
        addresses=addresses,
        post_up=post_up,
        pre_down=pre_down,
        default_endpoint=default_endpoint,
        peers={},
    )
    secret = WgWizardSecret.generate()
    WgWizard(config=config, secret=secret).dump(config_dir, interface, overwrite)


@main.command()
@interface_option
@config_dir_option
@option(
    "--name",
    "-n",
    prompt="Name of the peer",
    help="Name of the peer.",
)
@option(
    "--addresses",
    "-a",
    help="Interface.Address of the client.",
)
@option(
    "--client-allowed-ips",
    help="Peer.AllowedIPs of the client.",
)
@option(
    "--client-persistent-keepalive",
    prompt="Peer.PersistentKeepalive of the client",
    help="Peer.PersistentKeepalive of the client.",
    default=25,
)
@invert_qrcode_option
def add_peer(
    interface,
    config_dir,
    name,
    addresses,
    client_allowed_ips,
    client_persistent_keepalive,
    invert_qrcode,
):
    """Add a new client to a wg-wizard config."""
    from .core import WgWizard, WgWizardPeerConfig, export_wg_quick_config

    wg_wizard = WgWizard.from_dir(config_dir, interface)
    config = wg_wizard.config

    # check name
    if name in config.peers:
        raise ValueError("Peer name must be unique.")

    # get default addresses and client_allowed_ips and then ask the user
    if addresses is None:
        addresses = click.prompt(
            "Interface.Address of the client",
            default=str(config.find_next_available_interface()),
        )
    addresses = [ip.strip() for ip in addresses.split(",")]
    if client_allowed_ips is None:
        client_allowed_ips = click.prompt(
            "Peer.AllowedIPs of the client", default="0.0.0.0/0, ::/0"
        )
    client_allowed_ips = [ip.strip() for ip in client_allowed_ips.split(",")]

    # add a peer to the config
    peer_config = WgWizardPeerConfig(
        addresses=addresses,
        server_allowed_ips=addresses,
        client_allowed_ips=client_allowed_ips,
        client_persistent_keepalive=client_persistent_keepalive,
    )
    config.add_peer(name, peer_config)

    # generate secret
    wg_wizard.secret.generate_peer_secret(name)
    wg_wizard.dump(config_dir, interface)

    logger.info("Client's wg-quick config QR Code:")
    export_wg_quick_config(
        wg_wizard, text=False, qrcode=True, invert_qrcode=invert_qrcode, peer_name=name
    )


@main.command()
@interface_option
@config_dir_option
@option("--all", "-a", is_flag=True, help="Regenerate the whole secret file.")
@option("--server", "-s", is_flag=True, help="Regenerate the server secret.")
@option("--missing", "-m", is_flag=True, help="Only generate the missing peer secrets.")
@option(
    "--peer",
    "-p",
    multiple=True,
    help="Regenerate a specific peer secret. Can be specified multiple times",
)
@option(
    "--overwrite/--no-overwrite",
    is_flag=True,
    help="Whether to overwrite the keys without confirmation.",
)
def generate_keys(interface, config_dir, all, server, missing, peer, overwrite):
    """Generate or regenerate public, private and preshared keys."""
    from .core import WgWizard

    wg_wizard = WgWizard.from_dir(config_dir, interface)
    missing_peers = wg_wizard.generate_keys(
        regenerate_all=all,
        server=server,
        missing=missing,
        peers=peer,
        overwrite=overwrite,
    )
    if missing_peers:
        logger.info("Generated secret for missing peers %s.", missing_peers)
    if all or server or missing_peers or peer:
        wg_wizard.dump(config_dir, interface, config=False)
    else:
        logger.info("Nothing changed.")


@main.command()
@interface_option
@config_dir_option
def check(interface, config_dir):
    """Check whether the wg-wizard config is ready for export."""
    from .core import WgWizard

    WgWizard.from_dir(config_dir, interface).check_secret()


@main.command()
@interface_option
@config_dir_option
@option(
    "--text/--no-text",
    is_flag=True,
    default=True,
    help="Whether to output config text to the stdout.",
)
@option(
    "--qrcode/--no-qrcode",
    is_flag=True,
    help="Whether to output the QR Code to the stdout.",
)
@invert_qrcode_option
def export_server_config(interface, config_dir, text, qrcode, invert_qrcode):
    """Export a wg-quick server config."""
    from .core import export_wg_quick_config_from_files

    export_wg_quick_config_from_files(
        config_dir, interface, text, qrcode, invert_qrcode
    )


@main.command()
@interface_option
@config_dir_option
@option(
    "--name",
    "-n",
    prompt="Name of the peer",
    help="Name of the peer.",
)
@option(
    "--text/--no-text",
    is_flag=True,
    default=True,
    help="Whether to output config text to the stdout.",
)
@option(
    "--qrcode/--no-qrcode",
    is_flag=True,
    default=True,
    help="Whether to output the QR Code to the stdout.",
)
@invert_qrcode_option
def export_client_config(interface, config_dir, name, text, qrcode, invert_qrcode):
    """Export a wg-quick client config."""
    from .core import export_wg_quick_config_from_files

    export_wg_quick_config_from_files(
        config_dir, interface, text, qrcode, invert_qrcode, peer_name=name
    )



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['wg_wizard']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'pydantic>=1.9.2,<2.0.0',
 'qrcode>=7.3.1,<8.0.0',
 'ruamel.yaml>=0.17.21,<0.18.0']

entry_points = \
{'console_scripts': ['wg-wizard = wg_wizard.cli:main']}

setup_kwargs = {
    'name': 'wg-wizard',
    'version': '0.2.1',
    'description': 'Control all the WireGuard server and client configs in a centralized way.',
    'long_description': 'wg-wizard\n=========\n.. image:: https://github.com/ianlini/wg-wizard/actions/workflows/main.yml/badge.svg\n   :target: https://github.com/ianlini/wg-wizard/actions/workflows/main.yml\n.. image:: https://github.com/ianlini/wg-wizard/actions/workflows/release.yml/badge.svg\n   :target: https://github.com/ianlini/wg-wizard/actions/workflows/release.yml\n.. image:: https://img.shields.io/pypi/v/wg-wizard.svg\n   :target: https://pypi.org/project/wg-wizard/\n.. image:: https://img.shields.io/pypi/l/wg-wizard.svg\n   :target: https://github.com/ianlini/wg-wizard/blob/master/LICENSE\n.. image:: https://img.shields.io/github/stars/ianlini/wg-wizard.svg?style=social\n   :target: https://github.com/ianlini/wg-wizard\n\nwg-wizard can help you generate the configs for WireGuard interactively,\nand control all the server and client configs in a centralized way.\n\n.. contents::\n\nNetwork Architecture\n--------------------\n\nOur default network architecture is one\n`relay server <https://docs.sweeting.me/s/wireguard#Bounce-Server>`_\nwith multiple clients connecting to the VPN via the relay server.\n\nWe assume that most people need a network architecture like this.\nIf you want another architecture, you can still achieve some of them\nusing the advance configuration.\n\nWhy Do We Need wg-wizard\n------------------------\n\nTODO\n\nPrerequisite\n------------\n\n1. A server with Ubuntu, Debian or Raspberry Pi OS.\n   You can still use other Linux distributions,\n   but the following instructions might not be applicable.\n2. `Install Docker <https://docs.docker.com/engine/install/>`_ on your relay server.\n3. `Install Wireguard <https://www.wireguard.com/install/>`_ on your relay server.\n\nRun the Docker Container\n------------------------\n\nWe use Docker to disable the network to make the whole process secure.\n\n.. code-block:: sh\n\n   # go to the directory you want to put the configs\n   # assuming ~/wg-wizard here\n   export WG_WIZARD_CONFIG=\'~/wg-wizard\'\n   cd "${WG_WIZARD_CONFIG}"\n   docker run -it --rm --network=none --volume="$PWD":/workdir ianlini/wg-wizard\n\nInitialize the Configuration\n----------------------------\n\n.. code-block:: sh\n\n   # inside the Docker container\n   wg-wizard init\n\nFollow the instruction to create the config. Example output:\n\n.. code-block::\n\n   Interface name for WireGuard [wg0]:\n   Interface.ListenPort of the relay server [51820]:\n   Interface.Address of the relay server [192.168.10.1/24]:\n   The default endpoint in clients\' Peer.Endpoint configs (e.g., example.com:51280): example.com\n   If you want to allow the clients to access the internet via the relay server, you must provide the interface name you want to forward the internet traffic to. It\'s usually eth0 or wlan0. You can check it by executing `ip addr`. If you provide an interface name {interface}, the following rules will be added:\n   - iptables -A FORWARD -i %i -o {interface} -j ACCEPT\n   - iptables -A FORWARD -i {interface} -o %i -j ACCEPT\n   - iptables -t nat -A POSTROUTING -s {network} -o {interface} -j MASQUERADE\n   Interface name for connecting to the internet []: eth0\n   Do you want to allow the clients to connect with each other? If yes, a rule will be added: `iptables -A FORWARD -i %i -o %i -j ACCEPT` [Y/n]:\n   Do you want to allow the clients to connect to any IPs on the relay server? If no, only the IP of the WireGuard interface can be connected, that is, the following rules will be added:\n   - iptables -A INPUT -d {wg_server_interface_ip} -i %i -j ACCEPT\n   - iptables -A INPUT -i %i -j DROP\n   [y/N]:\n   [2022-10-12 18:41:34,190][INFO] wg_wizard.core: Writing config to /home/pi/pi-gateway/wireguard/wg0.yml\n   [2022-10-12 18:41:38,868][INFO] wg_wizard.core: Writing secret to /home/pi/pi-gateway/wireguard/wg0_secret.json\n\nNormally, you can use the default values for almost all of the options.\n\nIf you allow the internet access or allow the clients to connect with each other,\nyou also need to `enable IP forwarding <https://www.digitalocean.com/community/tutorials/how-to-set-up-wireguard-on-ubuntu-20-04#step-4-adjusting-the-wireguard-server-s-network-configuration>`_.\n\nFor convenience, in the following instructions,\nwe assume that your WireGuard interface name is ``wg0``:\n\n.. code-block:: sh\n\n   export WG_INTERFACE=wg0\n\nAdd a Peer\n----------\n\n.. code-block:: sh\n\n   # inside the Docker container\n   wg-wizard add-peer\n\nFollow the instruction to create the peer config. Example output:\n\n.. code-block::\n\n   Interface name for WireGuard [wg0]:\n   Name of the client: phone1\n   Peer.PersistentKeepalive of the client [25]:\n   Interface.Address of the client [192.168.10.2/32]:\n   Peer.AllowedIPs of the client [0.0.0.0/0, ::/0]:\n   [2022-10-04 16:40:01,337][INFO] wg_wizard.core: Writing config to /workdir/wg0.yml\n   [2022-10-04 16:40:01,358][INFO] wg_wizard.core: Writing secret to /workdir/wg0_secret.json\n   [2022-10-04 16:40:01,362][INFO] wg_wizard.cli: Client\'s wg-quick config QR Code:\n   ...\n\nNormally, you can use the default values for almost all of the options.\nIn the end, there will be a QR Code generated.\nYou can now use your `WireGuard app <https://www.wireguard.com/install/>`_\non your phone to scan the QR Code to import the config.\nIf your client doesn\'t support QR Code, you can use another command to generate the text:\n\n.. code-block:: sh\n\n   wg-wizard export-client-config --interface "${WG_INTERFACE}" --name phone1 --no-qrcode\n\nSet Up the WireGuard Server\n---------------------------\n\nPreparing:\n\n.. code-block:: sh\n\n   # on your relay server (outside the Docker container)\n   cd "${WG_WIZARD_CONFIG}"\n   export WG_INTERFACE=wg0  # replace wg0 with your interface name\n   (umask 077; sudo mkdir /etc/wireguard/)\n\nExporting server config:\n\n.. code-block:: sh\n\n   docker run --rm --network=none --volume="$PWD":/workdir ianlini/wg-wizard \\\n       wg-wizard export-server-config -i "${WG_INTERFACE}" \\\n       | sudo cp --backup /dev/stdin "/etc/wireguard/${WG_INTERFACE}.conf"\n\nIf you haven\'t enabled the service:\n\n.. code-block:: sh\n\n   # start the WireGuard server\n   sudo systemctl enable "wg-quick@${WG_INTERFACE}.service"\n   sudo systemctl start "wg-quick@${WG_INTERFACE}.service"\n\nNow you can turn on the WireGuard tunnel on your client (phone1),\nand it should work.\n\nIf the service is already running, you can check the config diff first:\n\n.. code-block:: sh\n\n   sudo diff "/etc/wireguard/${WG_INTERFACE}.conf~" "/etc/wireguard/${WG_INTERFACE}.conf"\n\nAfter confirming the changes, there are 2 ways to apply them.\n\n1. If you are not changing the wg-quick specific interface configs\n   (e.g., Address, DNS, MTU, Table, PreUp, PostUp, PreDown,\n   PostDown and SaveConfig),\n   you can reload the config without stopping the server:\n\n   .. code-block:: sh\n\n      sudo systemctl reload "wg-quick@${WG_INTERFACE}.service"\n\n2. Otherwise, you should restart the server:\n\n   .. code-block:: sh\n\n      sudo systemctl restart "wg-quick@${WG_INTERFACE}.service"\n\n\nTroubleshooting\n---------------\n\nRead the service log:\n\n.. code-block:: sh\n\n   journalctl -u "wg-quick@${WG_INTERFACE}.service" -f -n 1000\n\nEnable the kernel log:\n\n.. code-block:: sh\n\n   sudo modprobe wireguard\n   echo module wireguard +p | sudo tee /sys/kernel/debug/dynamic_debug/control\n\nRead the kernel log:\n\n.. code-block:: sh\n\n   journalctl -k -f -n 1000 | grep wireguard\n\nDebug iptables:\n\n.. code-block:: sh\n\n   # trace the ICMP packets from a WireGuard client\n   sudo iptables -t raw -A PREROUTING -i "${WG_INTERFACE}" -p icmp -j TRACE\n   # trace the incoming ICMP packets from the internet to a WireGuard client\n   sudo iptables -t mangle -A FORWARD -d 192.168.10.0/24 -p icmp -j TRACE\n\n.. warning::\n   Debugging iptables requires much more knowledge,\n   or you might generate large logs,\n   or even break the network of the whole machine.\n   However, it is highly possible that the generated configs don\'t work out-of-the-box.\n   It is the hard part when developing this tool because people will have different existing rules.\n   If you have a bad luck, you might need to spend some time understanding\n   the relationship between iptables and WireGuard.\n\nBuild the Docker Image from Scratch\n-----------------------------------\n\nTODO\n\n.. code-block:: sh\n\n   git clone ...\n   cd ...\n   docker build . -t ianlini/wg-wizard\n\nReferences\n----------\n\n- `Some Unofficial WireGuard Documentation <https://docs.sweeting.me/s/wireguard>`_\n- `How To Set Up WireGuard on Ubuntu 20.04 <https://www.digitalocean.com/community/tutorials/how-to-set-up-wireguard-on-ubuntu-20-04>`_\n- `How To Set Up WireGuard Firewall Rules in Linux <https://www.cyberciti.biz/faq/how-to-set-up-wireguard-firewall-rules-in-linux/>`_\n',
    'author': 'Ian Lin',
    'author_email': 'you@example.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ianlini/wg-wizard',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
